##############################################################################
#                                                                            #
#   Hyperspectral Irradiance Model for DiurnAll Light Levels (HEIMDALL)      #
#                                                                            #
# Midnight Sun to Polar Night: A model of seasonal light in the Barents Sea  #
#                    Connan-McGinty et al., 2022.                            #
#                                                                            #
##############################################################################   


import scipy.io
import matplotlib.pyplot as plt
import numpy as np
import math
import datetime
from scipy import integrate
from scipy.interpolate import CubicSpline
from scipy import interpolate
from datetime import date, timedelta
from datetime import datetime
import netCDF4
import time
import pandas as pd
import ephem
from skyfield import api
from skyfield import almanac
from skyfield.api import load
from skyfield.framelib import ecliptic_frame
import sys

#################################################################################################################
######################################## Import Solar Zenith Module  ############################################

# Specity path to Solar_Zenith_Calculator.py
path_2_solar_module = r'C:\Users\...'
sys.path.append(path_2_solar_module)

from Solar_Zenith_Calculator import zenith_calculator

#################################################################################################################
######################################### Load MAT files containing parameters ##################################

# Above surface Ed from Hydrolight and Spitschan et al. 2016

# Specity path to MAT files
MAT_Files_Path = r'C:\Users\...'
Spectral_Irradiance=scipy.io.loadmat(MAT_Files_Path + '\FINAL_Above_PAR_Ed_Spectrals.mat')
Ed_Diffuse=Spectral_Irradiance['Ed_Diffuse_Spitschan']                                     
Ed_Direct=Spectral_Irradiance['Ed_Direct_Hydro']       

# Load Lunar phase angles and functions from Miller and Turner 2009 
Lunar_Data = scipy.io.loadmat(MAT_Files_Path + '\Lunar_Phase_Functions_Albedos.mat')
phase_functions=Lunar_Data['Phase_Functions']
phase_angles=Lunar_Data['Phase_Angles']
spectral_albedos=Lunar_Data['Spectral_Albedos']                               

# Bio Optical Model Parameters from Kostakis et al. 2020.
Bio_Optical_Data = scipy.io.loadmat(MAT_Files_Path + '\Bio_Optical_Model_Parameters.mat')
A_b_B_b = Bio_Optical_Data['A_b_B_b']
A_bb_B_bb = Bio_Optical_Data['A_bb_B_bb']
Aph_Bph = Bio_Optical_Data['Aph_Bph']
H20_a_b_bb = Bio_Optical_Data['H20_a_b_bb']

#################################################################################################################
####################################### Intitialise output storage ##############################################

# Irradiance Values
Ed_Output={}
# Photon Counts
PAR_Output={}

#################################################################################################################
#################### Function to calculate Unix Timestamp for use in Solar_Zenith_Calculator.py #################

def unix_epoch(date_time_entry):
    dt =datetime.strptime(date_time_entry, "%Y-%m-%d  %H:%M:%S")                                           
    timestamp = int((dt - datetime(1970, 1, 1, 0, 0, 0)) / timedelta(seconds=1))   
    return timestamp

#################################################################################################################
##################### Function to calculate Unix Timestamp for converting instrument datetime ###################

def unix_epoch_instrument(date_time_entry,format_instrument_datetime):
    dt =datetime.strptime(date_time_entry, format_instrument_datetime)                                           
    timestamp = int((dt - datetime(1970, 1, 1, 0, 0, 0)) / timedelta(seconds=1))   
    return timestamp

#################################################################################################################
######################### Self written function for rounding lat/lon for cloud data #############################


def cloud_round(lat, lon):
    ERA5_lat = round(lat*4)/4
    ERA5_lon = round(lon*4)/4
    return ERA5_lat, ERA5_lon

#################################################################################################################
########################### Calculate starting time for Copernicus Cloud Cover Data #############################

def unix2cloud_date(ts):
    data1 = datetime.utcfromtimestamp(ts)           # unix timestamp to datetime object
    data2 = datetime(2011,1,1,0,0,0)                  # starting date for cloud cover data
    
    diff = data1 - data2
    
    days, seconds = diff.days, diff.seconds
    hours = (days * 24 + seconds // 3600) +1       # add 1 hour for NETCDF Data
    return hours

#################################################################################################################
######################################### Lunar Model Functions #################################################

# Linear inter for phase angles - assumption full moon phase_func = 1, new moon phase_func = 0
def phase_function_interp(wavelength,lunar_phase):
    wavelengths=np.arange(405,705,10)
    interp_2D=scipy.interpolate.interp2d(wavelengths,phase_angles,phase_functions, kind='cubic')
    return interp_2D(wavelength,lunar_phase)

# Spectral Albedo Interpolator for wavelengths outside of binned values 
def spectral_albedo_interp(wavelength):
    wavelengths=np.arange(400,705,100)
    interp_1D=scipy.interpolate.interp1d(wavelengths,spectral_albedos, kind='linear')
    return interp_1D(wavelength)

def Lunar_Zenith_Calculator(lat,lon,timestamp):    
    # Ensures position provided is relative to Earth observor                                
    gatech = ephem.Observer()     
    # Must be input as string - otherwise ephem will read float as radians                  
    gatech.lat, gatech.lon = str(lat), str(lon)     
    #Convert Unix timestamp to datetime readable by ephem
    gatech.date = datetime.utcfromtimestamp(timestamp).strftime('%Y-%m-%d %H:%M')     
    sun, moon = ephem.Sun(), ephem.Moon()
    sun.compute(gatech)
    moon.compute(gatech)
    
    alt=("%s" % (moon.alt))
     
    if (alt[0]=='-') and (alt[2]==':'):
        lunar_elevation_angle=(float(alt[1])+float(alt[3:5])/60)  * -1
    elif (alt[0]=='-') and (alt[3]==':'):
        lunar_elevation_angle=(float(alt[1:3])+float(alt[4:6])/60 ) * -1
    elif alt[1]==':':
        lunar_elevation_angle=float(alt[0])+float(alt[2:4].zfill(2))/60
    else:
        lunar_elevation_angle=float(alt[0:2])+float(alt[3:5].zfill(2))/60
       
    lunar_zenith= 90 - lunar_elevation_angle    

    return lunar_zenith

def Lunar_Phase_Distance_Calculator(timestamp):
    ts = load.timescale()
    UTC_Timestamp=datetime.utcfromtimestamp(timestamp).strftime('%Y-%m-%d %H:%M')
    
    t = ts.utc(int(UTC_Timestamp[0:4]), int(UTC_Timestamp[5:7]), int(UTC_Timestamp[8:10]), int(UTC_Timestamp[11:13]), int(UTC_Timestamp[14:16]))
    
    
    eph = load('de421.bsp')
    sun, moon, earth = eph['sun'], eph['moon'], eph['earth']
    
    e = earth.at(t)
    _, slon, _ = e.observe(sun).apparent().frame_latlon(ecliptic_frame)
    _, mlon, m_distance = e.observe(moon).apparent().frame_latlon(ecliptic_frame)
    phase = (mlon.degrees - slon.degrees) % 360.0
    
    # Convention in this calculator is 0° New Moon, 90° First Quarter, 180° Full Moon, and 270° Last Quarter.
    # Miller and Turner 2009 [Phase Function Data] define terms as 180° New Moon to 0° Full Moon back to 180°.
    
    phase_angle = '{0:.2f}'.format(phase)                                                   
    lunar_phase = float('{0:.2f}'.format(abs(180 - float(phase_angle))))
    lunar_distance = float(m_distance.au) 
    
    return lunar_distance, lunar_phase

#################################################################################################################
################################### Above Surface Ed Calculation Functions ######################################


def Diffuse_Ed_Interp(Cloud,Zenith):
    
    interp_type='linear'
    
    if Zenith>106:
        Zenith = 106
    hydro_cloud_cover_values=np.arange(0,101,10) 
    Zenith_Values_Hydro=np.arange(0,81,10) 
    Zenith_Values_Spitschan = Spectral_Irradiance['Spitshcan_zeniths'][0,:]
    Zenith_Values_Spitschan = np.append(Zenith_Values_Spitschan,106)
                                 
    hydro_zenith_values=list(np.concatenate([Zenith_Values_Hydro,Zenith_Values_Spitschan]))      
    cs2D=scipy.interpolate.interp2d(hydro_cloud_cover_values, hydro_zenith_values, Ed_Diffuse[ind,:,:], kind=interp_type)
    return cs2D(Cloud,Zenith)



def Direct_Ed_Interp(Cloud,Zenith):
    
    interp_type='linear'
    
    if Zenith>=90:
        Zenith = 90
        
    hydro_cloud_cover_values=np.arange(0,101,10) 
    hydro_zenith_values=np.arange(0,91,10)    
    cs2D=scipy.interpolate.interp2d(hydro_cloud_cover_values, hydro_zenith_values, Ed_Direct[ind,:,:], kind=interp_type)
    return cs2D(Cloud,Zenith)

    

#################################################################################################################
################################## Surface Transmission Function ###############################################  


def Fresnel_Direct_Transmission_Factor(zenith):
    if zenith>=90:
        return 0
    else:
       n1=1.00
       n2=1.34
       Rs=((n1*(np.cos((zenith*np.pi)/180)) - n2*np.sqrt(1-((n1/n2)*np.sin((zenith*np.pi)/180))**2))/(n1*(np.cos((zenith*np.pi)/180)) + n2*np.sqrt(1-((n1/n2)*np.sin((zenith*np.pi)/180))**2)))**2
       Rp=((n1*np.sqrt(1-((n1/n2)*np.sin((zenith*np.pi)/180))**2)-n2*np.cos((zenith*np.pi)/180))/(n1*np.sqrt(1-((n1/n2)*np.sin((zenith*np.pi)/180))**2)+n2*np.cos((zenith*np.pi)/180)))**2
       Reflection=(Rs+Rp)/2
       Transmission=(1-Reflection)
       return Transmission

def Diffuse_Transmission_Factor():
    Reflect = 0.0288
    Transmission=(1-Reflect)     
    return Transmission
    
#################################################################################################################
##################################### Attenuation Function ######################################################

def SP_Attenuation(zenith, a, b, bb, z, PAR_0_direct, PAR_0_diffuse):
    
    def Surface_Mean_Cosine(zenith):
        sin_theta_w = (np.sin(np.deg2rad(zenith)))/1.34
        theta_w=np.arcsin(sin_theta_w)
        mu_w = np.cos(theta_w)
        return mu_w
    
    
   
    mu_w = Surface_Mean_Cosine(zenith)
    K_direct = (1/mu_w) * (a + bb)
    
    K_diffuse =  (1/0.83) * (a + bb)

    
    PAR_Direct =  PAR_0_direct * np.exp(-K_direct * z)
    PAR_Diffuse = PAR_0_diffuse * np.exp(-K_diffuse * z)
    
        
    return PAR_Direct , PAR_Diffuse

#################################################################################################################
############################## Bio Optical Model - Kostakis et al., 2020 ########################################

def a_TOTAL(wavelength,Chl,Sal):
    
    ### Water Absorption ###
    a_w = H20_a_b_bb[list(H20_a_b_bb[:,0]).index(wavelength),1]
    
    ### CDOM Absorption ###
    a_CDOM_440 = -0.012*Sal + 0.464
    a_CDOM = a_CDOM_440 * np.exp(-0.0168*(wavelength-440))
    
    ### Phytoplankton Absorption ###
    A_ph = Aph_Bph[list(Aph_Bph[:,0]).index(wavelength),1]
    B_ph = Aph_Bph[list(Aph_Bph[:,0]).index(wavelength),2]
    a_ph = A_ph*(Chl**(1-B_ph))
    
    ### NAP Absorption ###
    a_nap_440 = 0.0124*(Chl**0.724)
    a_nap = a_nap_440 * np.exp(-0.011*(wavelength-440))
    
    ### TOTAL ###
    a_total = a_w + a_CDOM + a_ph + a_nap
    return a_total


def b_TOTAL(wavelength,Chl):
    
    ### Water Scattering ###
    b_w = H20_a_b_bb[list(H20_a_b_bb[:,0]).index(wavelength),2]
    
    ### Particulate Matter Scattering ###
    A_b = A_b_B_b[list(A_b_B_b[:,0]).index(wavelength),1]
    B_b = A_b_B_b[list(A_b_B_b[:,0]).index(wavelength),2]
    b_p = A_b*(Chl**(B_b))
    
    ### TOTAL ###
    b_total= b_w + b_p
    return b_total


def bb_TOTAL(wavelength,Chl):
    
    ### Water Backscattering ###
    bb_w = H20_a_b_bb[list(H20_a_b_bb[:,0]).index(wavelength),3]
    
    ### Particulate Matter Backscattering ###
    A_bb = A_bb_B_bb[list(A_bb_B_bb[:,0]).index(wavelength),1]
    B_bb = A_bb_B_bb[list(A_bb_B_bb[:,0]).index(wavelength),2]
    bb_p = A_bb*(Chl**(B_bb))
    
    ### TOTAL ###
    bb_total = bb_w + bb_p
    return bb_total


#################################################################################################################
########################################### Input  Timestamps ###################################################

def define_date_time():
    global time_stamps
    global time_resolution
    global output_times
    global time_resolution_steps
    
    import_timestamps=input('Are instrument timestamps being imported? Y/N: ')
    if import_timestamps == 'Y':
        import_timestamps = True    
    if import_timestamps == 'N':
        import_timestamps = False
    
    if import_timestamps == True:
        unix_check=input('Are timestamps in unix format? Y/N: ')
        if unix_check == 'Y':
            time_stamps_input = str(input('Please import the array containing instrument UTC timestamp values: '))
            exec("global time_stamps_out; time_stamps_out = {}".format(time_stamps_input))
            time_stamps = time_stamps_out 
            
        if unix_check == 'N':
            # See https://docs.python.org/3/library/datetime.html for Python datetime formats
            format_instrument_datetime = str(input('Please enter the format of datetime data, e.g. %Y-%m-%d  %H:%M:%S: '))
            Datetime_Input = str(input('Please import the array containing instrument UTC timestamp values: '))
            exec("global Date_Times; Date_Times = {}".format(Datetime_Input))
            
            time_stamps = [unix_epoch_instrument(t,format_instrument_datetime) for t in Date_Times] 
            
        time_resolution_steps = len(time_stamps)
        
    if import_timestamps == False:
        start_date = input('Enter a start date/time in YYYY-MM-DD HH:MM:SS format:')
        start_dt =datetime.strptime(start_date, "%Y-%m-%d  %H:%M:%S")
        
        end_date = input('Enter an end date/time in YYYY-MM-DD HH:MM:SS format:')
        end_dt =datetime.strptime(end_date, "%Y-%m-%d  %H:%M:%S")
        
        difference = end_dt - start_dt
        days, seconds = difference.days, difference.seconds
        total_seconds_timespan = days* 86400 + seconds
        time_resolution = int(float(input('Please enter required time resolution in minutes: ')) * 60)
        
        
        # Number of time steps that irradiance will be calculated 
        time_resolution_steps= int(np.ceil(total_seconds_timespan / time_resolution)) + 1
        time_stamps = np.arange(unix_epoch(start_date),unix_epoch(end_date)+time_resolution,time_resolution)
        output_times = [datetime.utcfromtimestamp(unix_timestamp) for unix_timestamp in time_stamps]
        

#################################################################################################################
####################################### Input Water Constituents ################################################

def constituent_inputs():
    global Chl
    global Sal
    
    
    water_column_homog=input('Are the water column constituents homogeneous for all time steps? Y/N: ')
    if water_column_homog == 'Y':
        water_column_homog = True    
    if water_column_homog == 'N':
        water_column_homog = False
    
    if water_column_homog == True:
        Chl_Input = float(input('Please enter a chlorophyll concentration value: '))
        Sal_Input = float(input('Please enter a salinity value: '))
        Chl = Chl_Input * np.ones(np.shape(spatial_temporal_matrix))
        Sal = Sal_Input * np.ones(np.shape(spatial_temporal_matrix))
    
    if water_column_homog == False:
        Chl_Location = input('Please import the array containing chlorophyll concentration values: ')
        exec("global Chl; Chl = {}".format(Chl_Location))
        Sal_Location = input('Please import the array containing salinity values: ')
        exec("global Sal; Sal = {}".format(Sal_Location))

#################################################################################################################

##################################### Input Variable Check ######################################################

def var_check():
    while((np.shape(Chl)!=np.shape(spatial_temporal_matrix)) or ((np.shape(Sal)!=np.shape(spatial_temporal_matrix)))):
        input_change_option = int(input('Constituent input size does not match depth or time resolution. \nOption 1: Change water constituent input. \nOption 2: Change datetime input.  '))
        if input_change_option==1: 
            constituent_inputs()
        elif input_change_option==2: 
            define_date_time()    
            
#################################################################################################################  
#################################### Input variables from user ##################################################


lat = float(input('Please enter a latitude: '))
lon = float(input('Please enter a longitude: '))
cloud_option=int(input('Please choose a Cloud Cover Option: \n0:Clear Sky \n1:100% Cloud Cover \n2:ERA5 Cloud Cover Data: '))
bottom_depth = int(input('Please enter a bottom boundary depth to nearest 1m: '))
depth_resolution = int(input('Please enter required depth resolution to nearest 1m: '))
depths = np.arange(depth_resolution,bottom_depth+1,depth_resolution)
depth_steps = np.insert(depths, 0 ,0, axis=None)
define_date_time()

spatial_temporal_matrix =  np.transpose([depths] * time_resolution_steps)

constituent_inputs()

var_check()

#################################################################################################################
########################################### Set up path to cloud data ###########################################

imported_cloud_data=np.zeros(len(time_stamps))
if cloud_option==1:
        imported_cloud_data=np.ones(len(time_stamps))


# Specify path to ERA5 cloud data and filename. Comment lines 430-433 out
# if not using satellite cloud cover data.
path = '...'
cloud_filename = '...'

copernicus_cloud_cover = netCDF4.Dataset(path+cloud_filename)

#################################################################################################################
##############################################    Begin Model     ###############################################
         
start = time.time()
print('\nImporting Cloud Cover Data ...')
if cloud_option==2:
    
    hour_indices=np.zeros(time_resolution_steps)
    
    
    for i in range(time_resolution_steps):
        hour_indices[i] = unix2cloud_date(time_stamps[i])
    
    
    
    # Use lat/lon determined by cloud_round() function. ERA5 satellite data needs lat/lon at nearest 0.25°.
    for i in range(time_resolution_steps):
        imported_cloud_data[i]=round(copernicus_cloud_cover['tcc'][hour_indices[i],
                                list(copernicus_cloud_cover['latitude'][:]).index(cloud_round(lat, lon)[0]),list(copernicus_cloud_cover['longitude'][:]).index(cloud_round(lat, lon)[1])],2)
        
Cloud=imported_cloud_data*100


print('Cloud Cover Data Imported.')   


surface_area_moon = np.pi * (1737400) **2           #Surface Area Moon in m^(2)
One_Astro_Unit = 149597870700                       # 1 AU in m
wavelengths=np.arange(405,705,10)

total_diffuse_spectral_instantaneous_Ed=np.zeros((time_resolution_steps,30))
total_direct_spectral_instantaneous_Ed=np.ones((time_resolution_steps,30))

solar_diffuse_spectral_instantaneous_Ed=np.zeros((time_resolution_steps,30))
solar_direct_spectral_instantaneous_Ed=np.ones((time_resolution_steps,30))

lunar_diffuse_spectral_instantaneous_Ed=np.zeros((time_resolution_steps,30))
lunar_direct_spectral_instantaneous_Ed=np.ones((time_resolution_steps,30))

print('\nCalculating above surface spectral irradiance ...')

for i in range(time_resolution_steps):
    
    solar_zenith_angle=zenith_calculator(lat,lon,str(int(time_stamps[i]))) 
    lunar_zenith_angle = Lunar_Zenith_Calculator(lat,lon,time_stamps[i])                   
    lunar_distance = Lunar_Phase_Distance_Calculator(time_stamps[i])[0] * One_Astro_Unit
    lunar_solid_angle = surface_area_moon / (lunar_distance **2)
    
    for ind in range(30):
    
        solar_diffuse_spectral_instantaneous_Ed[i,ind]=Diffuse_Ed_Interp(Cloud[i],solar_zenith_angle) 
        solar_direct_spectral_instantaneous_Ed[i,ind]=Direct_Ed_Interp(Cloud[i],solar_zenith_angle)
        
        if solar_zenith_angle>90:
            solar_direct_spectral_instantaneous_Ed[i,ind]=0
        
        
        spectral_albedo=spectral_albedo_interp(wavelengths[ind])
        phase_function_factor=phase_function_interp(wavelengths[ind],(Lunar_Phase_Distance_Calculator(time_stamps[i])[1]))
        
        lunar_diffuse_spectral_instantaneous_Ed[i,ind] = Diffuse_Ed_Interp(Cloud[i],lunar_zenith_angle) * spectral_albedo * phase_function_factor * lunar_solid_angle   
        lunar_direct_spectral_instantaneous_Ed[i,ind] = Direct_Ed_Interp(Cloud[i],lunar_zenith_angle) * spectral_albedo * phase_function_factor * lunar_solid_angle   
        
        if lunar_zenith_angle>90:
            lunar_direct_spectral_instantaneous_Ed[i,ind]=0
        # Below condition stops posibility of galactic component being accounted for twice.
        if lunar_zenith_angle>=106:
            lunar_diffuse_spectral_instantaneous_Ed[i,ind]=0 
            
        total_diffuse_spectral_instantaneous_Ed[i,ind] =  solar_diffuse_spectral_instantaneous_Ed[i,ind] + lunar_diffuse_spectral_instantaneous_Ed[i,ind]
        total_direct_spectral_instantaneous_Ed[i,ind] =  solar_direct_spectral_instantaneous_Ed[i,ind] + lunar_direct_spectral_instantaneous_Ed[i,ind]
        
print('Above surface spectral irradiance calculation complete.')   
         
#################################################################################################################
###################################   Convert Spectral Ed to Spectral PAR   #####################################
            
h=6.626e-34                               #Planck's Constant
c=3e8                                     #Speed of Light
photon_wavelengths=np.arange(405,705,10)*1e-9
photon_energies=[(h*c)/x for x in photon_wavelengths]          
            
direct_spectral_instantaneous_PAR=np.zeros((time_resolution_steps,30))            
diffuse_spectral_instantaneous_PAR=np.zeros((time_resolution_steps,30)) 

print('\nConverting spectral irradiance to PAR units ...')
for j in range(30):
    direct_spectral_instantaneous_PAR[:,j]=10*total_direct_spectral_instantaneous_Ed[:,j]/photon_energies[j]
    diffuse_spectral_instantaneous_PAR[:,j]=10*total_diffuse_spectral_instantaneous_Ed[:,j]/photon_energies[j]

print('Conversion to PAR units complete.')    
#################################################################################################################            
##########################################   Surface Transmission   #############################################


below_surface_direct_spectral_PAR=np.zeros((time_resolution_steps,30))
below_surface_diffuse_spectral_PAR=np.zeros((time_resolution_steps,30))

below_surface_direct_spectral_Ed=np.zeros((time_resolution_steps,30))
below_surface_diffuse_spectral_Ed=np.zeros((time_resolution_steps,30))


print('\nCalculating surface transmission ...')
for ind in range(time_resolution_steps):
    
    
    solar_zen = zenith_calculator(lat,lon,str(time_stamps[i]))
    lunar_zen = Lunar_Zenith_Calculator(lat,lon,time_stamps[i])
    if solar_zen >=90:
       zenith_angle = lunar_zen
    else:
        zenith_angle = solar_zen
    
    below_surface_direct_spectral_PAR[ind,:]=direct_spectral_instantaneous_PAR[ind,:]*Fresnel_Direct_Transmission_Factor(zenith_angle)
    below_surface_diffuse_spectral_PAR[ind,:]=diffuse_spectral_instantaneous_PAR[ind,:]*Diffuse_Transmission_Factor()
    
    below_surface_direct_spectral_Ed[ind,:]=total_direct_spectral_instantaneous_Ed[ind,:]*Fresnel_Direct_Transmission_Factor(zenith_angle)
    below_surface_diffuse_spectral_Ed[ind,:]=total_diffuse_spectral_instantaneous_Ed[ind,:]*Diffuse_Transmission_Factor()

print('Surface transmission calculation complete.')
#################################################################################################################
########################################## Attenuation Calculation ##############################################

spectral_underwater_Ed=np.zeros((time_resolution_steps,len(depth_steps),30))
spectral_underwater_PAR=np.zeros((time_resolution_steps,len(depth_steps),30))


print('\nCalculating water column attenuation ...')
for i in range(time_resolution_steps):
    
    solar_zen = zenith_calculator(lat,lon,str(time_stamps[i]))
    lunar_zen = Lunar_Zenith_Calculator(lat,lon,time_stamps[i])
    if solar_zen >=90:
       zenith_angle = lunar_zen
    else:
        zenith_angle = solar_zen
       
    for m in range(30):
        
        PAR_0_diffuse = below_surface_diffuse_spectral_PAR[i,m]
        PAR_0_direct = below_surface_direct_spectral_PAR[i,m]
        
        Ed_0_diffuse = below_surface_diffuse_spectral_Ed[i,m]
        Ed_0_direct = below_surface_direct_spectral_Ed[i,m]
        
        spectral_underwater_PAR[i,0,m] = below_surface_diffuse_spectral_PAR[i,m] + below_surface_direct_spectral_PAR[i,m]
        spectral_underwater_Ed[i,0,m] = below_surface_diffuse_spectral_Ed[i,m] + below_surface_direct_spectral_Ed[i,m]
        
        for n in np.arange(0, len(depths),1):
            
            atten_depth = depth_steps[n+1] - depth_steps[n]
    
            a = a_TOTAL(int(photon_wavelengths[m]/1e-9),Chl[n,i],Sal[n,i])
            b = b_TOTAL(int(photon_wavelengths[m]/1e-9),Chl[n,i])
            bb = bb_TOTAL(int(photon_wavelengths[m]/1e-9),Chl[n,i])
        
        
            
            PAR_direct, PAR_diffuse = SP_Attenuation(zenith_angle, a, b , bb, atten_depth, PAR_0_direct, PAR_0_diffuse)
            
            spectral_underwater_PAR[i,n+1,m] = (PAR_direct + PAR_diffuse)
            
            Ed_direct, Ed_diffuse = SP_Attenuation(zenith_angle, a, b , bb,  atten_depth, Ed_0_direct, Ed_0_diffuse )
            
            spectral_underwater_Ed[i,n+1,m] = (Ed_direct + Ed_diffuse)
            
            PAR_0_direct = PAR_direct
            PAR_0_diffuse = PAR_diffuse
            Ed_0_direct = Ed_direct
            Ed_0_diffuse = Ed_diffuse

           
            
print('Water column attenuation calculation complete.')            

#################################################################################################################
##################################### -------- Save Files -------- ##############################################


output_times = [datetime.utcfromtimestamp(unix_timestamp) for unix_timestamp in time_stamps]

# Outputs are 3-D Array of shape (Timestamps, Depths, Wavelenth)
# PAR Units of photons m-2 s-1, integrated over 10nm waveband.
PAR_Output['Spectral_Underwater_PAR'] = spectral_underwater_PAR
PAR_Output['Time_Steps'] = output_times
PAR_Output['Output_Depths'] = depth_steps
PAR_Output['Wavelengths'] = wavelengths


# Outputs are 3-D Array of shape (Timestamps, Depths, Wavelenth)
# Ed Units of W m-2 nm-1  (Bandwidth 10nm)
Ed_Output['Spectral_Underwater_Ed'] = spectral_underwater_Ed
Ed_Output['Time_Steps'] = output_times
Ed_Output['Output_Depths'] = depth_steps
Ed_Output['Wavelengths'] = wavelengths


# User can create directory here to save dictionaries to...
#
#
#

#################################################################################################################

print('\n\n_______COMPLETE_______ \nTotal time in seconds:', round(time.time()-start,2))

############################################### END #############################################################